# coding=utf-8

import types
import collections

try:
    from urlparse import urlparse
except ImportError:
    from urllib.parse import urlparse


class Pagination(object):

    """Pagination using for navigation on website."""

    def __init__(self, xpath=None, host=None, function=None):
        """Override initialization instance.

        :param xpath (optional): `str` xpath for search next link.
        :param host (optional): `str` host URL.
        :param function (optional): `function` function for get next link.
        """
        self.__xpath = xpath
        self.__host = host
        self.__function = function
        self.__last = None

    def __iter__(self):
        """Override instance iteration."""
        if self.__function is not None:
            result = self.__function()
            if isinstance(result, types.GeneratorType):
                return result
            elif isinstance(result, collections.Iterable):
                return iter(result)
            else:
                return self

    def next(self, page=None):
        """Return next link.

        :param page (optional): `lxml.Element` instance.
        :returns: `str` link.
        """
        if self.__function is not None:
            url = self.__function(page)
        elif self.__xpath is not None and self.__host is not None:
            try:
                url = page.xpath(self.__xpath)[0]
                url = urlparse(self.__host + url).geturl()
            except IndexError:
                return None
        else:
            raise AttributeError(
                'Cannot call `.next()` as no `xpath=` and `host=` or '
                '`function=` keywords argument was passed when instantiating '
                'the pagination instance.'
            )

        if self.__last == url:
            raise StopIteration()
        else:
            self.__last = url
            return url

    def __next__(self):
        """Return next link.

        :param page (optional): `lxml.Element` instance.
        :returns: `str` link.
        """
        return self.next()



# coding=utf-8

import requests
from lxml import html

class Crawler(object):

    """Crawler parse page use items as rules. May be nested."""

    def __init__(self, url, items, crawlers=None, session=None):
        """Override initialization instance.

        :param url: `str` URL for page.
        :param items: `dict` items.
        :param crawlers (optional): `dict` crawlers.
        :param session (optional): `requests.Session` instance.
        """
        self.__url = url
        self.__items = items or {}
        self.__crawlers = crawlers or {}
        self.__session = session or requests.Session()
        self.__data = {}

    @property
    def data(self):
        """The data property.

        :returns: `list` data.
        """
        return self.__data

    def crawl(self):
        """Crawl page.

        :returns: `dict` data.
        """
        page = html.fromstring(
            self.__session.get(self.__url, verify=False).content
        )

        for name, item in self.__items.items():
            self.__data[name] = item.parse(page)

        for name, crawler in self.__crawlers.items():
            self.__data[name] = crawler.crawl()

        return self.data



# coding=utf-8


class Field(object):

    """Field is minimum structure unit. Contains certain value."""

    def __init__(self, xpath=None, function=None, postprocessing=None, to=str):
        """Override initialization instance.

        :param xpath (optional): `str` xpath for extracting value from page.
        :param function (optional): `function` function for getting value.
        :param postprocessing (optional): `function` postprocessing function.
        :param to (optional): `type` to representation as.
        """
        self.__xpath = xpath
        self.__function = function
        self.__postprocessing = postprocessing
        self.__to = to

    def parse(self, page):
        """Parse value from page.

        :param page: `lxml.Element` instance.
        :returns: value.
        """
        if self.__xpath is not None:
            value = page.xpath(self.__xpath)
        elif self.__function is not None:
            value = self.__function(page)
        else:
            raise AttributeError(
                'Cannot call `.search()` as no `value=`, `xpath=` or '
                '`function=` keywords argument was passed when instantiating '
                'the field instance.'
            )

        if self.__to not in (list, tuple):
            try:
                value = value[0]
            except IndexError:
                value = None

        if value is not None and self.__postprocessing is not None:
            value = self.__postprocessing(value)

        return value


class Item(object):

    """Item is aggregate of fields. Items can be nested."""

    def __init__(self, fields, xpath='//html'):
        """Override initialization instance.

        :param fields: `dict` fields.
        :param xpath (optional): `str` xpath for extracting blocks from page.
        """
        for field in fields.values():
            assert issubclass(field.__class__, (Field, Item)), (
                '`fields` must be `Field` or `Item` instances.'
            )

        self.__xpath = xpath
        self.__fields = fields

    def parse(self, page):
        """Parse process.

        :param page: `lxml.Element` instance.
        :returns: `list` data.
        """
        data = []

        for block in page.xpath(self.__xpath):
            item = {}
            for name, field in self.__fields.items():
                item[name] = field.parse(block)

            data.append(item)

        return data



# coding=utf-8

import configobj


class Settings(object):

    """Class for keep settings."""

    def __init__(self, configspec=None, configuration=None):
        """Override initialization instance.

        :param configspec (optional): `dict` configspec for read from file.
        :param configuration (optional): `dict` configuration.
        """
        self.__configspec = configspec
        self.__configuration = configuration

    def __getattr__(self, attribute):
        """Override getting not existing attribute."""
        assert not (self.__configspec and not self.__configuration), (
            'Cannot access to settings before call `.load_from_file` method.'
        )
        assert self.__configspec or self.__configuration, (
            'Cannot access to settings as no `configuration=` or '
            '`configspec=` keywords argument was not passed when instantiating'
            ' the settings instance.'
        )

        value = self.__configuration[attribute]
        if isinstance(value, dict):
            return self.__class__(configuration=value)
        return value

    def load_from_file(self, filename):
        """
        Load settings from configuration file.

        :param filename: `str` configuration filename.
        :returns: `dict` of settings.
        """
        self.__configuration = configobj.ConfigObj(
            filename, configspec=self.__configspec
        ).dict()

    def load_from_dict(self, instance):
        """
        Load settings from `dict`.

        :param instance: `dict` settings.
        :returns: `dict` of settings.
        """
        self.__configuration = instance

    def create_configuration_file(self, filename):
        """
        Create configuration file.

        :param filename: `str` path to write configuration.
        """
        config = configobj.ConfigObj(infile=filename)
        config.update(self.__configspec)
        config.write()



# coding=utf-8

import requests

from metacrawler.settings import Settings


class Handler(object):

    """Handler control crawling process."""

    def __init__(self, crawlers):
        """Override initialization object.

        :param crawlers: `dict` crawlers.
        """
        self.__crawlers = crawlers

        self.session = requests.Session()
        self.settings = Settings()
        self.data = {}

    def before(self):
        """Any actions before start."""
        pass

    def start(self):
        """Start crawling."""
        self.before()

        for name, crawler in self.__crawlers.items():
            self.data[name] = crawler.crawl()
        return self.data



# coding=utf-8

from os import path
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

VERSION = '0.0.1'

BASE_DIR = path.abspath(path.dirname(__file__))

packages = [
    'metacrawler'
]

requires = [
    'configobj',
    'lxml',
    'requests',
]

with open(path.join(BASE_DIR, 'README.rst')) as f:
    long_description = f.read()

setup(
    name='metacrawler',
    version=VERSION,
    packages=packages,
    install_requires=requires,
    description='Create your own crawler.',
    long_description=long_description,
    author='Vitalii Maslov',
    author_email='me@pyvim.com',
    url='https://github.com/pyvim/metacrawler',
    download_url='https://github.com/pyvim/metacrawler/tarball/master',
    license='MIT',
    keywords='parser,crawler',
    classifiers=[
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 3',
    ],
)
